@@ -10,10 +10,11 @@ import warnings
 
 import numpy as np
 import pandas as pd
-from numpy.core._exceptions import _ArrayMemoryError
 from numpy.typing import NDArray
 from simplejpeg import decode_jpeg, encode_jpeg
 
+from dhcomp.composite import _greedy_composite
+from typing import Optional
 
 class ClassHeaders(NamedTuple):
     class_number: int
@@ -35,7 +36,8 @@ class CrasHeader(NamedTuple):
     id: str  # starts with "CoreLog Linescan ".   If it starts with "CoreLog Linescan 1." then it supports compression (otherwise ignore ctype).
     ns: int  # image width in pixels
     nl: int  # image height in lines
-    nb: int  # number of bands (1 or 3  but always 3 for HyLogger 1 / 2 / 3)
+    nb: int  # number of bands (1 or 3  but always 3 for HyLogger 1 / 2 / 3)- added mir tsg for HyLogger 4
+    nb: int  # number of bands (1 or 3  but always 3 for HyLogger 1 / 2 / 3)- added mir tsg for HyLogger 4
     org: int  # interleave (1=BIL  2=BIP  and compressed rasters are always BIP while uncompressed ones are always BIL)
     dtype: int  # datatype (unused  always byte)
     specny: int  # number of linescan lines per dataset sample
@@ -46,7 +48,9 @@ class CrasHeader(NamedTuple):
     nchunks: int  # number of compressed image chunks (jpeg compression)
     csize32_obs: int  # size in bytes of comressed image data (OBSOLETE - not used anywhere any more.   However it will be set in old linescan rasters so I cant easily recycle it.   Also  there are some compressed rasters out there that are >4GB in size)
     ntrays: int  # number of trays (number of tray-table records after the image data)
-    nsections: int  # number of sections (number of section-table records after the image data)
+    nsections: (
+        int  # number of sections (number of section-table records after the image data)
+    )
     finerep: int  # chip-mode datasets - number of spectral measurements per chip bucket (and theres one image frame per bucket)
     jpqual: int  # jpeg quality factor  0..100 (jpeg compression)
 
@@ -60,7 +64,9 @@ class TrayInfo(NamedTuple):
 
 
 class SectionInfo(NamedTuple):
-    utlengthmm: float  # untrimmed length of imagery in mm (could be less than the tray's)
+    utlengthmm: (
+        float  # untrimmed length of imagery in mm (could be less than the tray's)
+    )
     startmm: float  # start position (along scan) in mm
     endmm: float  # end position in mm
     trimwidthmm: float  # active (section-overlap-corrected) image width in mm
@@ -97,9 +103,10 @@ class Spectra:
 @dataclass
 class TSG:
     nir: Spectra
-    tir: Spectra
-    cras: Cras
-    lidar: Union[NDArray, None]
+    tir: Optional[Spectra]
+    mir: Optional[Spectra]
+    cras: Optional[Cras]
+    lidar: Optional[NDArray]
 
     def __repr__(self) -> str:
         tsg_info: str = "This is a TSG file"
@@ -113,6 +120,8 @@ class FilePairs:
     nir_bip: Union[Path, None] = None
     tir_tsg: Union[Path, None] = None
     tir_bip: Union[Path, None] = None
+    mir_tsg: Union[Path, None] = None
+    mir_bip: Union[Path, None] = None
     lidar: Union[Path, None] = None
     cras: Union[Path, None] = None
 
@@ -168,6 +177,14 @@ class FilePairs:
             valid = True
         return valid
 
+    def valid_mir(self) -> bool:
+        result = self._get_bip_tsg_pair("mir")
+        if result is None:
+            valid = False
+        else:
+            valid = True
+        return valid
+
     def valid_lidar(self) -> bool:
         result = self._get_lidar()
         if result is None:
@@ -185,7 +202,9 @@ class FilePairs:
         return valid
 
 
-def read_cras(filename: Union[str, Path], big_file=False) -> Cras:
+def read_cras(
+    filename: Union[str, Path], backing_file: Union[str, Path, None] = None
+) -> Cras:
     """Read a cras file
 
     Args:
@@ -197,7 +216,6 @@ def read_cras(filename: Union[str, Path], big_file=False) -> Cras:
     section_info_format: str = "4f3i"
     tray_info_format: str = "3f2i"
     head_format: str = "20s2I8h4I2h"
-
     with open(filename, "rb") as file:
         # using memory mapping
 
@@ -232,15 +250,39 @@ def read_cras(filename: Union[str, Path], big_file=False) -> Cras:
         # we currently are reading the entire cras.bip file
         # which can cause issues due to memory allocation  well handle that case
         # with some error handling here where we quit while we are ahead
-        try:
-            cras = np.zeros((header.nl, header.ns, header.nb), dtype=np.uint8)
-            array_ok = True
-        except _ArrayMemoryError:
-            print(
-                "This file is too big to fit inmemory set big_file=True to dump to disk"
+        # assume that the drive backing the file has enough space to store the files
+        # assume that there is enough ram to load the cras
+        array_ok: bool = True
+        cras: Union[NDArray, zarr.core.Array]  # type: ignore
+        if backing_file is not None:
+            import zarr
+            # the big file flag decompresses the jpg data into a zarr array
+            # with zarr it is important to ensure that you set the chunks appropriately
+            # this means that you are aligning the output chunk size to the input chunk size
+            outcras: Path
+            if isinstance(backing_file, str):
+                outcras = Path(backing_file)
+            else:
+                outcras = backing_file
+
+            cras = zarr.open(
+                str(outcras),
+                mode="w",
+                shape=(header.nl, header.ns, header.nb),
+                chunks=(header.chunksize, header.ns, header.nb),
+                dtype=np.uint8,
             )
-            array_ok = False
-            cras = np.zeros(1, dtype=np.uint8)
+
+        else:
+            try:
+                cras = np.zeros((header.nl, header.ns, header.nb), dtype=np.uint8)
+                array_ok = True
+            except numpy.core._exceptions._ArrayMemoryError:
+                print(
+                    "This file is too big to fit in memory set big_file=True to dump to disk"
+                )
+                array_ok = False
+                cras = np.zeros(1, dtype=np.uint8)
         # if the array fits into memory then proceed to decode the .jpgs
         # using the chunk_offset_array to correctly index to the right location
         # TODO: it might be worthwhile to modify this code to manage the case
@@ -260,7 +302,7 @@ def read_cras(filename: Union[str, Path], big_file=False) -> Cras:
                 img = decode_jpeg(chunk, colorspace="BGR")
                 # reverse the channels
                 # and flip the image upsidedown
-                np_image = np.flipud(img)
+                np_image = np.flipud(img)  # type: ignore
                 nr = np_image.shape[0]
                 cras[curpos : (curpos + nr), :, :] = np_image
                 curpos = curpos + nr
@@ -284,13 +326,81 @@ def read_cras(filename: Union[str, Path], big_file=False) -> Cras:
         for i in range(header.nsections):
             bytes = file.read(28)
             section.append(SectionInfo(*struct.unpack(section_info_format, bytes)))
+    if backing_file is not None:
+        pd.DataFrame(section).to_csv(backing_file.with_suffix(".section"))
+        pd.DataFrame(tray).to_csv(backing_file.with_suffix(".tray"))
 
     output = Cras(cras, tray, section)
     return output
 
 
+def composite_spectra(spectra: Spectra, length: int = 4) -> Spectra:
+    """
+    Composites the spectra to a target interval length in mm
+    uses dhcomp's greedy_composite internally
+    keeps all the information pertaining to depth registration but looses
+    all the other parameters in the scalars so there is none of the extra information that
+    tsg normally provides
+
+    """
+    secdist = spectra.sampleheaders["X"].astype(float).map(np.round).copy()
+    sections = spectra.sampleheaders["T"] + spectra.sampleheaders["L"]
+    sections = sections.map({n: i for i, n in enumerate(sections.unique())})
+    new_intervals = sections.values * 0
+    int_intervals = sections.values * 0
+
+    int_interval = 0
+    for i in sections.astype(int).unique():
+        sidx = sections == i
+        section = sections[sidx]
+        cut_points = _greedy_composite(secdist[sidx].values, length)
+        tmp_section = section.values * 0
+        for n, (fr, to) in enumerate(zip(cut_points[:-1], cut_points[1:])):
+            tmpidx = (secdist[sidx].values >= fr) & (secdist[sidx].values <= to)
+            tmp_section[tmpidx] = n
+        new_intervals[sidx.values] = tmp_section
+        int_intervals[sidx.values] = tmp_section + int_interval
+        int_interval += tmp_section.max() + 1
+
+    comp_spectra = np.ones((int_interval, spectra.spectra.shape[1]))
+    for i in range(int_interval):
+        idx = int_intervals == i
+        comp_spectra[i, :] = spectra.spectra[idx].mean(0)
+    spectra.spectra = comp_spectra
+
+    spectra.sampleheaders["new_intervals"] = new_intervals
+    spectra.sampleheaders["Section"] = sections
+    spectra.sampleheaders[["new_intervals"]] + spectra.sampleheaders[["Section"]]
+    new_depths = (
+        spectra.sampleheaders[
+            ["Section", "sample", "D", "T", "P", "X", "L", "new_intervals"]
+        ]
+        .astype(float)
+        .groupby(["Section", "new_intervals"])
+        .max()
+        .reset_index()
+    )
+    spectra.scalars = spectra.scalars.copy()
+    spectra.scalars["new_intervals"] = new_intervals
+    spectra.scalars["Section"] = sections
+    new_scalay = (
+        spectra.scalars[["Section", "new_intervals", "SecDist (mm)"]]
+        .astype(float)
+        .groupby(["Section", "new_intervals"])
+        .max()
+        .reset_index()
+    )
+    spectra.sampleheaders = new_depths
+    spectra.scalars = new_scalay
+    spectra.sampleheaders
+    return spectra
+
+
 def extract_chips(
-    filename: Union[str, Path], outfolder: Union[str, Path], spectra: Spectra
+    filename: Union[str, Path],
+    outfolder: Union[str, Path],
+    spectra: Spectra,
+    centre_cut: bool = True,
 ):
     if isinstance(outfolder, str):
         outfolder = Path(outfolder)
@@ -360,14 +470,34 @@ def extract_chips(
     # but such are the vagaries of ML
     # I totally assume that these headers always exist in the scalars
 
-    section_array = (
-        spectra.sampleheaders["L"].astype(int).values - 1
-    )  # subtract 1 so that we are 0 indexed
+    # here we will do some trickery to reindex unique combinations of tray and line
+    # by dropping the duplicates combinations of T and L and then reindexing
+    # merging that index to the original file and then iterating over that.
+
+    # test to see if the scalars have either ['T', 'L'] which should indicate a diamond drill hole
+    # if ['Tray', 'Section'] exist then it is likely that we are dealing with chip data
+    headers: list[str]
+    if spectra.sampleheaders.columns.isin(["Tray", "Section"]).sum() == 2:
+        headers = ["Tray", "Section"]
+    elif spectra.sampleheaders.columns.isin(["T", "L"]).sum() == 2:
+        headers = ["T", "L"]
+    elif spectra.sampleheaders.columns.isin(["sample"]).sum() == 1:
+        headers = ["sample"]
+    tmp_headers: pd.DataFrame = (
+        spectra.sampleheaders[headers].drop_duplicates().reset_index()
+    )
+    # get the index and set it's value to the column called index
+    tmp_headers["index"] = tmp_headers.index
+    # extract the index and use that as the section array
+    section_array: NDArray = spectra.sampleheaders.merge(tmp_headers)["index"].values
     sample_length = spectra.scalars["SecDist (mm)"].diff()
     # this is only na for the first sample
     idx_sample_na = (sample_length.isna()) | (sample_length < 0)
     sample_length[idx_sample_na] = spectra.scalars["SecDist (mm)"][idx_sample_na]
     # pd is slow for lots of accesses
+
+    # aggregate the spectra to modulo
+
     sample_array: NDArray = sample_length.values
     curchunk: int = 0
     cursample: int = 0
@@ -396,7 +526,6 @@ def extract_chips(
         yres = sec.utlengthmm / sec.nlines
         # allocate the section
         working = np.zeros((sec.nlines, header.ns, header.nb), dtype="uint8")
-        curpos = 0
         # if we've dropped any information into the leading bin
         # dump it out into for into the working array
         if np.any(leading_bin):
@@ -409,6 +538,7 @@ def extract_chips(
                 (header.chunksize, header.ns, header.nb), dtype="uint8"
             )
         # you need to monitor the processed lines to maintain this loop
+
         while (curchunk * header.chunksize - processed_lines) < sec.nlines:
             total_offset = chunk_offset_array[curchunk] + 4 * (header.nchunks + 1) + 64
             chunksize_in_bytes = (
@@ -421,15 +551,26 @@ def extract_chips(
             nr = np_image.shape[0]
             end_pos = curpos + nr
 
-            if end_pos <= sec.nlines:
+            if end_pos < sec.nlines:
                 working[curpos:end_pos, :, :] = np_image
             elif end_pos > sec.nlines:
                 nextra = end_pos - sec.nlines
                 end_pos = sec.nlines
                 end_np = nr - nextra
-                working[curpos:end_pos, :, :] = np_image[0:end_np]
+                if nextra == 0:
+                    end_np = nr
+                # case when curpos is gt manage this by appending all data in this chunk to the leading bin
+                if curpos < end_pos:
+                    working[curpos:end_pos, :, :] = np_image[0:end_np]
+                    # put the remaining information into leading bin
+                    leading_bin[0:nextra, :, :] = np_image[end_np:nr]
+                else:
+                    # put the remaining information into leading bin
+                    leading_bin[0:nr, :, :] = np_image
+
+            elif end_pos == sec.nlines:
                 # put the remaining information into leading bin
-                leading_bin[0:nextra, :, :] = np_image[end_np:nr]
+                leading_bin[0:nr, :, :] = np_image
 
             curpos = curpos + nr
             # increment the chunk
@@ -437,184 +578,242 @@ def extract_chips(
 
         # book keeping the processed lines
         processed_lines += sec.nlines
-
         idx_section = section_array == i
         im_cuts = np.floor(sample_array[idx_section] / yres).astype(int)
         cut_array = np.concatenate([[0], np.cumsum(im_cuts).ravel()])
-
         n_cuts = len(cut_array)
         for j in range(n_cuts - 1):
-            current_image = working[cut_array[j] : cut_array[j + 1]]
-            tmp_file = "{}.jpg".format(cursample)
-            outfile = outfolder.joinpath(tmp_file)
-            outjpg = encode_jpeg(current_image)
-            with open(outfile, "wb") as tmpf:
-                tmpf.write(outjpg)
-            cursample += 1
+            if cut_array[j + 1] <= sec.nlines:
+                current_image = working[cut_array[j] : cut_array[j + 1]]
+                if centre_cut:
+                    # cut the image square where y i.e. depth is equal to scan width
+                    mid_point = int(current_image.shape[1] / 2)
+                    n_half = current_image.shape[0] // 2
+                    current_image = current_image[
+                        :, (mid_point - n_half) : (mid_point + n_half), :
+                    ].copy()
+
+                tmp_file = "{}.jpg".format(cursample)
+                outfile = outfolder.joinpath(tmp_file)
+                outjpg = encode_jpeg(current_image)
+                with open(outfile, "wb") as tmpf:
+                    tmpf.write(outjpg)
+                cursample += 1
 
 
-def extract_chips(
-    filename: Union[str, Path], outfolder: Union[str, Path], spectra: Spectra
-):
-    if isinstance(outfolder, str):
-        outfolder = Path(outfolder)
-
-    if not outfolder.exists():
-        outfolder.mkdir()
+def generate_chips(
+    filename: Union[str, Path],
+    spectra: Spectra,
+    centre_cut: bool = True,
+    batch_size: int = 256,
+) -> tuple[NDArray]:
+    """
+    creates an generator that generates the image tiles the last batch is not guaranteed to be the target size
+    avoids having to write to folder, useful for processing files without having to first write to disk.
+    """
 
     section_info_format: str = "4f3i"
     tray_info_format: str = "3f2i"
     head_format: str = "20s2I8h4I2h"
-    file = open(filename, "rb")
-    # read the 64 byte header from the .cras file
-    bytes = file.read(64)
-    # create the header information
-    header = CrasHeader(*struct.unpack(head_format, bytes))
+    with open(filename, "rb") as file:
+        # read the 64 byte header from the .cras file
+        bytes = file.read(64)
+        # create the header information
+        header = CrasHeader(*struct.unpack(head_format, bytes))
 
-    # Create the chunk_offset_array
-    # which determines which point of the file to enter to read the .jpg image
+        # Create the chunk_offset_array
+        # which determines which point of the file to enter to read the .jpg image
 
-    file.seek(64)
-    b = file.read(4 * (header.nchunks + 1))
-    chunk_offset_array = np.ndarray((header.nchunks + 1), np.uint32, b)
+        file.seek(64)
+        b = file.read(4 * (header.nchunks + 1))
+        chunk_offset_array = np.ndarray((header.nchunks + 1), np.uint32, b)
 
-    # check for the existance of sections or  trays
-    # if they exist we are going to skip ahead and import them first
-    # as we are going to use them to section the images to a per spectrum basis
-    # so the cras file uses compressed jpg chunks of approximately fixed dimension
-    # so we need to use the section and tray information to calculate the correct
-    # image size that matches the spectra
-    # we will set up an array that we will use to accumulate the images into
-    # then as each image is accumulated we dump it to disk and call name it the sample name
-    # it is likely that we need two accumulation arrays the first as a bin to hold the images
-    # as they are read to disk and the second to hold the image that we are going to export
-    if header.nsections > 0 or header.ntrays > 0:
-        # the tray info section if it exists should start after the last image
-        # the section info and if there is a tray info section then it should be after the tray info section
-        info_table_start = (
-            64
-            + (header.nchunks + 1) * 4
-            + chunk_offset_array[header.nchunks]
-            - chunk_offset_array[0]
+        # check for the existance of sections or  trays
+        # if they exist we are going to skip ahead and import them first
+        # as we are going to use them to section the images to a per spectrum basis
+        # so the cras file uses compressed jpg chunks of approximately fixed dimension
+        # so we need to use the section and tray information to calculate the correct
+        # image size that matches the spectra
+        # we will set up an array that we will use to accumulate the images into
+        # then as each image is accumulated we dump it to disk and call name it the sample name
+        # it is likely that we need two accumulation arrays the first as a bin to hold the images
+        # as they are read to disk and the second to hold the image that we are going to export
+        if header.nsections > 0 or header.ntrays > 0:
+            # the tray info section if it exists should start after the last image
+            # the section info and if there is a tray info section then it should be after the tray info section
+            info_table_start = (
+                64
+                + (header.nchunks + 1) * 4
+                + chunk_offset_array[header.nchunks]
+                - chunk_offset_array[0]
+            )
+            file.seek(info_table_start)
+
+            tray: "list[TrayInfo]" = []
+            for i in range(header.ntrays):
+                bytes = file.read(20)
+                tray.append(TrayInfo(*struct.unpack(tray_info_format, bytes)))
+
+            section: "list[SectionInfo]" = []
+            for i in range(header.nsections):
+                bytes = file.read(28)
+                section.append(SectionInfo(*struct.unpack(section_info_format, bytes)))
+
+        # it seems to be best to allocate memory for each of the sections if there are multiple sections we
+        # empty the array and create a new one of the correct dimension
+        # on third thoughts we will precalculate which chunks are going to which section because we know that
+        # then loop over sets of chunks dumping to disk incrementally.
+        # at this stage I'm not sure it will work on drill core
+        # no the header contains the chunk dimensions
+        # loop over the section
+        # it seems that you need to have the sample header information from the
+        # nir/tir/mir spectra we use nir because it should always be there
+        # once we have that information we are going to caculate the number of pixels required
+        # in the y direction that represent a single spectrum and the option will also be to dump
+        # all the spectra to disk named as H_SAMPLE in a subfolder which will take an impressive amount of space
+        # but such are the vagaries of ML
+        # I totally assume that these headers always exist in the scalars
+
+        # here we will do some trickery to reindex unique combinations of tray and line
+        # by dropping the duplicates combinations of T and L and then reindexing
+        # merging that index to the original file and then iterating over that.
+
+        # test to see if the scalars have either ['T', 'L'] which should indicate a diamond drill hole
+        # if ['Tray', 'Section'] exist then it is likely that we are dealing with chip data
+        headers: list[str]
+        if spectra.sampleheaders.columns.isin(["Tray", "Section"]).sum() == 2:
+            headers = ["Tray", "Section"]
+        elif spectra.sampleheaders.columns.isin(["T", "L"]).sum() == 2:
+            headers = ["T", "L"]
+        elif spectra.sampleheaders.columns.isin(["sample"]).sum() == 1:
+            headers = ["sample"]
+        tmp_headers: pd.DataFrame = (
+            spectra.sampleheaders[headers].drop_duplicates().reset_index()
         )
-        file.seek(info_table_start)
+        # get the index and set it's value to the column called index
+        tmp_headers["index"] = tmp_headers.index
+        # extract the index and use that as the section array
+        section_array: NDArray = spectra.sampleheaders.merge(tmp_headers)[
+            "index"
+        ].values
+        sample_length = spectra.scalars["SecDist (mm)"].diff()
+        # this is only na for the first sample
+        idx_sample_na = (sample_length.isna()) | (sample_length < 0)
+        sample_length[idx_sample_na] = spectra.scalars["SecDist (mm)"][idx_sample_na]
+        # pd is slow for lots of accesses
 
-        tray: "list[TrayInfo]" = []
-        for i in range(header.ntrays):
-            bytes = file.read(20)
-            tray.append(TrayInfo(*struct.unpack(tray_info_format, bytes)))
+        # aggregate the spectra to modulo
 
-        section: "list[SectionInfo]" = []
-        for i in range(header.nsections):
-            bytes = file.read(28)
-            section.append(SectionInfo(*struct.unpack(section_info_format, bytes)))
+        sample_array: NDArray = sample_length.values
+        curchunk: int = 0
+        cursample: int = 0
+        processed_lines: int = 0
+        yres: float
+        working: NDArray[np.uint8]
+        curpos: int = 0
+        nr: int
+        pos_fill: NDArray[np.int32]
+        idx_bin_fill: NDArray[np.bool8]
+        leading_bin: NDArray[np.uint8] = np.zeros(
+            (header.chunksize, header.ns, header.nb), dtype="uint8"
+        )
+        total_offset: int
+        chunksize_in_bytes: int
+        np_image: NDArray[np.uint8]
+        end_pos: int
+        nextra: int
+        end_np: int
+        idx_section: NDArray[np.bool8]
+        cut_array: NDArray[np.int32]
+        n_cuts: int
+        cutouts: list[NDArray[np.uint8]] = []
+        biter: int = 0
+        for i, sec in enumerate(section):
+            # pixel resolution
+            yres = sec.utlengthmm / sec.nlines
+            # allocate the section
+            working = np.zeros((sec.nlines, header.ns, header.nb), dtype="uint8")
+            # if we've dropped any information into the leading bin
+            # dump it out into for into the working array
+            if np.any(leading_bin):
+                idx_bin_fill = np.all(np.any(leading_bin, 1), 1)
+                pos_fill = np.where(idx_bin_fill)[0]
+                working[pos_fill] = leading_bin[pos_fill]
+                curpos = pos_fill[-1] + 1
+                # empty the leading bin
+                leading_bin = np.zeros(
+                    (header.chunksize, header.ns, header.nb), dtype="uint8"
+                )
+            # you need to monitor the processed lines to maintain this loop
 
-    # it seems to be best to allocate memory for each of the sections if there are multiple sections we
-    # empty the array and create a new one of the correct dimension
-    # on third thoughts we will precalculate which chunks are going to which section because we know that
-    # then loop over sets of chunks dumping to disk incrementally.
-    # at this stage I'm not sure it will work on drill core
-    # no the header contains the chunk dimensions
-    # loop over the section
-    # it seems that you need to have the sample header information from the
-    # nir/tir spectra we use nir because it should always be there
-    # once we have that information we are going to caculate the number of pixels required
-    # in the y direction that represent a single spectrum and the option will also be to dump
-    # all the spectra to disk named as H_SAMPLE in a subfolder which will take an impressive amount of space
-    # but such are the vagaries of ML
-    # I totally assume that these headers always exist in the scalars
+            while (curchunk * header.chunksize - processed_lines) < sec.nlines:
+                total_offset = (
+                    chunk_offset_array[curchunk] + 4 * (header.nchunks + 1) + 64
+                )
+                chunksize_in_bytes = (
+                    chunk_offset_array[curchunk + 1] - chunk_offset_array[curchunk]
+                )
+                file.seek(total_offset)
+                chunk = file.read(chunksize_in_bytes)
+                np_image = decode_jpeg(chunk, colorspace="BGR")
+                np_image = np.flipud(np_image)
+                nr = np_image.shape[0]
+                end_pos = curpos + nr
 
-    section_array = (
-        spectra.sampleheaders["L"].astype(int).values - 1
-    )  # subtract 1 so that we are 0 indexed
-    sample_length = spectra.scalars["SecDist (mm)"].diff()
-    # this is only na for the first sample
-    idx_sample_na = (sample_length.isna()) | (sample_length < 0)
-    sample_length[idx_sample_na] = spectra.scalars["SecDist (mm)"][idx_sample_na]
-    # pd is slow for lots of accesses
-    sample_array: NDArray = sample_length.values
-    curchunk: int = 0
-    cursample: int = 0
-    processed_lines: int = 0
-    yres: float
-    working: NDArray[np.uint8]
-    curpos: int = 0
-    nr: int
-    pos_fill: NDArray[np.int32]
-    idx_bin_fill: NDArray[np.bool8]
-    leading_bin: NDArray[np.uint8] = np.zeros(
-        (header.chunksize, header.ns, header.nb), dtype="uint8"
-    )
-    total_offset: int
-    chunksize_in_bytes: int
-    np_image: NDArray[np.uint8]
-    end_pos: int
-    nextra: int
-    end_np: int
-    idx_section: NDArray[np.bool8]
-    cut_array: NDArray[np.int32]
-    n_cuts: int
-    tmp_file: str
-    for i, sec in enumerate(section):
-        # pixel resolution
-        yres = sec.utlengthmm / sec.nlines
-        # allocate the section
-        working = np.zeros((sec.nlines, header.ns, header.nb), dtype="uint8")
-        curpos = 0
-        # if we've dropped any information into the leading bin
-        # dump it out into for into the working array
-        if np.any(leading_bin):
-            idx_bin_fill = np.all(np.any(leading_bin, 1), 1)
-            pos_fill = np.where(idx_bin_fill)[0]
-            working[pos_fill] = leading_bin[pos_fill]
-            curpos = pos_fill[-1] + 1
-            # empty the leading bin
-            leading_bin = np.zeros(
-                (header.chunksize, header.ns, header.nb), dtype="uint8"
-            )
-        # you need to monitor the processed lines to maintain this loop
-        while (curchunk * header.chunksize - processed_lines) < sec.nlines:
-            total_offset = chunk_offset_array[curchunk] + 4 * (header.nchunks + 1) + 64
-            chunksize_in_bytes = (
-                chunk_offset_array[curchunk + 1] - chunk_offset_array[curchunk]
-            )
-            file.seek(total_offset)
-            chunk = file.read(chunksize_in_bytes)
-            np_image = decode_jpeg(chunk, colorspace="BGR")
-            np_image = np.flipud(np_image)
-            nr = np_image.shape[0]
-            end_pos = curpos + nr
+                if end_pos < sec.nlines:
+                    working[curpos:end_pos, :, :] = np_image
+                elif end_pos > sec.nlines:
+                    nextra = end_pos - sec.nlines
+                    end_pos = sec.nlines
+                    end_np = nr - nextra
+                    if nextra == 0:
+                        end_np = nr
+                    # case when curpos is gt manage this by appending all data in this chunk to the leading bin
+                    if curpos < end_pos:
+                        working[curpos:end_pos, :, :] = np_image[0:end_np]
+                        # put the remaining information into leading bin
+                        leading_bin[0:nextra, :, :] = np_image[end_np:nr]
+                    else:
+                        # put the remaining information into leading bin
+                        leading_bin[0:nr, :, :] = np_image
 
-            if end_pos <= sec.nlines:
-                working[curpos:end_pos, :, :] = np_image
-            elif end_pos > sec.nlines:
-                nextra = end_pos - sec.nlines
-                end_pos = sec.nlines
-                end_np = nr - nextra
-                working[curpos:end_pos, :, :] = np_image[0:end_np]
-                # put the remaining information into leading bin
-                leading_bin[0:nextra, :, :] = np_image[end_np:nr]
+                elif end_pos == sec.nlines:
+                    # put the remaining information into leading bin
+                    leading_bin[0:nr, :, :] = np_image
 
-            curpos = curpos + nr
-            # increment the chunk
-            curchunk += 1
+                curpos = curpos + nr
+                # increment the chunk
+                curchunk += 1
 
-        # book keeping the processed lines
-        processed_lines += sec.nlines
+            # book keeping the processed lines
+            processed_lines += sec.nlines
+            idx_section = section_array == i
+            im_cuts = np.floor(sample_array[idx_section] / yres).astype(int)
+            cut_array = np.concatenate([[0], np.cumsum(im_cuts).ravel()])
+            n_cuts = len(cut_array)
 
-        idx_section = section_array == i
-        im_cuts = np.floor(sample_array[idx_section] / yres).astype(int)
-        cut_array = np.concatenate([[0], np.cumsum(im_cuts).ravel()])
+            for j in range(n_cuts - 1):
+                if cut_array[j + 1] <= sec.nlines:
+                    current_image = working[cut_array[j] : cut_array[j + 1]]
+                    if centre_cut:
+                        # cut the image square where y i.e. depth is equal to scan width
+                        mid_point = int(current_image.shape[1] / 2)
+                        n_half = current_image.shape[0] // 2
+                        current_image = current_image[
+                            :, (mid_point - n_half) : (mid_point + n_half), :
+                        ].copy()
+                    cutouts.append(current_image)
+                    cursample += 1
 
-        n_cuts = len(cut_array)
-        for j in range(n_cuts - 1):
-            current_image = working[cut_array[j] : cut_array[j + 1]]
-            tmp_file = "{}.jpg".format(cursample)
-            outfile = outfolder.joinpath(tmp_file)
-            outjpg = encode_jpeg(current_image)
-            with open(outfile, "wb") as tmpf:
-                tmpf.write(outjpg)
-            cursample += 1
+                # if (biter >= batch_size) or (i >=header.nsections) or (len(cutouts)>=batch_size):
+                if (cursample % batch_size == 0) or (i >= header.nsections):
+                    biter = 0
+                    tupout = tuple(cutouts)
+                    cutouts = []
+                    yield tupout
+                else:
+                    biter += 1
+        yield tuple(cutouts)
 
 
 def _read_tsg_file(filename: Union[str, Path]) -> "list[str]":
@@ -904,7 +1103,10 @@ def _parse_tsg(
 
 
 def _parse_scalars(
-    scalars: NDArray, classes: "list[ClassHeaders]", bandheaders: "list[BandHeaders]"
+    scalars: NDArray,
+    classes: "list[ClassHeaders]",
+    bandheaders: "list[BandHeaders]",
+    nodata: int = -1,
 ) -> pd.DataFrame:
     """
     function to map the scalars to a pandas data frame with names and
@@ -916,7 +1118,13 @@ def _parse_scalars(
         # handle flag 13 that has a path to plsscalars
         if i.flag == 2:
             if i.class_number > 0:
-                bv = band_value.astype(int)
+                # replace missing values with `nodata (defaulting to -1)
+                # Note: the integer shouldn't exist in the `classes` keys
+                bv = np.where(
+                    np.isclose(band_value, np.finfo("float32").min),
+                    -1,
+                    band_value,
+                ).astype(int)
                 tn = classes[i.class_number].map_ints(bv)
                 tmp_series.append(pd.DataFrame(tn, columns=[i.name]))
             else:
@@ -956,6 +1164,7 @@ def read_package(
     read_cras_file: bool = False,
     extract_cras: bool = False,
     imageoutput: Union[str, None] = None,
+    backing_file: Union[Path, str, None] = None,
 ) -> TSG:
     # convert string to Path because we are wanting to use Pathlib objects to manage the folder structure
     if isinstance(foldername, str):
@@ -970,7 +1179,7 @@ def read_package(
     # we will parse the lidar height data because we can
 
     # process here is to map the files that we need together
-    # tir and nir files
+    # tir and nir files and mir in hylogger 4
     #
     # deal the files to the type
 
@@ -990,6 +1199,13 @@ def read_package(
         elif f.name.endswith("tsg_tir.bip"):
             setattr(file_pairs, "tir_bip", f)
 
+      
+        elif f.name.endswith("tsg_mir.tsg"):
+            setattr(file_pairs, "mir_tsg", f)
+
+        elif f.name.endswith("tsg_mir.bip"):
+            setattr(file_pairs, "mir_bip", f)
+
         elif f.name.endswith("tsg_cras.bip"):
             setattr(file_pairs, "cras", f)
 
@@ -1002,9 +1218,10 @@ def read_package(
     # for the nir/swir and then tir
     # read nir/swir
     nir: Spectra
-    tir: Spectra
-    lidar: Union[NDArray, None]
-    cras: Cras
+    mir: Optional[Spectra]
+    tir: Optional[Spectra]
+    lidar: Optional[NDArray]
+    cras: Optional[Cras] = None
 
     if file_pairs.valid_nir():
         nir = read_tsg_bip_pair(file_pairs.nir_tsg, file_pairs.nir_bip, "nir")
@@ -1015,6 +1232,12 @@ def read_package(
         tir = read_tsg_bip_pair(file_pairs.tir_tsg, file_pairs.tir_bip, "tir")
     else:
         tir = Spectra
+
+    if file_pairs.valid_mir():
+        mir = read_tsg_bip_pair(file_pairs.mir_tsg, file_pairs.mir_bip, "mir")
+    else:
+        mir = Spectra
+
     if file_pairs.valid_lidar():
         lidar = read_hires_dat(file_pairs.lidar)
     else:
@@ -1032,14 +1255,12 @@ def read_package(
 
             extract_chips(file_pairs.cras, imageoutput, nir)
             cras = Cras
-        else:
-            cras = read_cras(file_pairs.cras)
-
+        cras = read_cras(file_pairs.cras, backing_file)
     else:
         cras = Cras
 
-    return TSG(nir, tir, cras, lidar)
-
+    return TSG(nir,tir,mir, cras, lidar)
+    
 
 if __name__ == "main":
     foldername = "data/RC_hyperspectral_geochem"
